# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Dict, List, Tuple, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import fft2
from atommic.collections.common.parts.utils import expand_op
from atommic.collections.multitask.rs.nn.base import BaseMRIReconstructionSegmentationModel
from atommic.collections.multitask.rs.nn.mtlrs_base.mtlrs_block import MTLRSBlock
from atommic.core.classes.common import typecheck

__all__ = ["MTLRS"]


class MTLRS(BaseMRIReconstructionSegmentationModel):
    """Implementation of the Multi-Task Learning for MRI Reconstruction and Segmentation (MTLRS) model, as presented
    in [Karkalousos2023]_.

    References
    ----------
    .. [Karkalousos2023] Karkalousos, D., Išgum, I., Marquering, H., Caan, M. W. A., (2023). MultiTask Learning for
        accelerated-MRI Reconstruction and Segmentation of Brain Lesions in Multiple Sclerosis. In Proceedings of
        Machine Learning Research (Vol. 078).

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`MTLRS`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration object.
        trainer : Trainer, optional
            PyTorch Lightning trainer object. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.reconstruction_module_recurrent_filters = cfg_dict.get("reconstruction_module_recurrent_filters")
        self.reconstruction_module_time_steps = cfg_dict.get("reconstruction_module_time_steps")
        self.reconstruction_module_num_cascades = cfg_dict.get("reconstruction_module_num_cascades")
        self.reconstruction_module_accumulate_predictions = cfg_dict.get(
            "reconstruction_module_accumulate_predictions"
        )
        conv_dim = cfg_dict.get("reconstruction_module_conv_dim")
        reconstruction_module_params = {
            "num_cascades": self.reconstruction_module_num_cascades,
            "time_steps": self.reconstruction_module_time_steps,
            "no_dc": cfg_dict.get("reconstruction_module_no_dc"),
            "keep_prediction": cfg_dict.get("reconstruction_module_keep_prediction"),
            "dimensionality": cfg_dict.get("reconstruction_module_dimensionality"),
            "recurrent_layer": cfg_dict.get("reconstruction_module_recurrent_layer"),
            "conv_filters": cfg_dict.get("reconstruction_module_conv_filters"),
            "conv_kernels": cfg_dict.get("reconstruction_module_conv_kernels"),
            "conv_dilations": cfg_dict.get("reconstruction_module_conv_dilations"),
            "conv_bias": cfg_dict.get("reconstruction_module_conv_bias"),
            "recurrent_filters": self.reconstruction_module_recurrent_filters,
            "recurrent_kernels": cfg_dict.get("reconstruction_module_recurrent_kernels"),
            "recurrent_dilations": cfg_dict.get("reconstruction_module_recurrent_dilations"),
            "recurrent_bias": cfg_dict.get("reconstruction_module_recurrent_bias"),
            "depth": cfg_dict.get("reconstruction_module_depth"),
            "conv_dim": conv_dim,
            "pretrained": cfg_dict.get("pretrained"),
            "accumulate_predictions": self.reconstruction_module_accumulate_predictions,
        }

        self.segmentation_module_output_channels = cfg_dict.get("segmentation_module_output_channels", 2)
        segmentation_module_params = {
            "segmentation_module": cfg_dict.get("segmentation_module"),
            "output_channels": self.segmentation_module_output_channels,
            "channels": cfg_dict.get("segmentation_module_channels", 64),
            "pooling_layers": cfg_dict.get("segmentation_module_pooling_layers", 2),
            "dropout": cfg_dict.get("segmentation_module_dropout", 0.0),
            "temporal_kernel": cfg_dict.get("segmentation_module_temporal_kernel", 1),
            "activation": cfg_dict.get("segmentation_module_activation", "elu"),
            "bias": cfg_dict.get("segmentation_module_bias", False),
            "conv_dim": conv_dim,
        }

        self.coil_dim = cfg_dict.get("coil_dim", 1)
        self.consecutive_slices = cfg_dict.get("consecutive_slices", 1)

        self.rs_cascades = cfg_dict.get("joint_reconstruction_segmentation_module_cascades", 1)
        self.rs_module = torch.nn.ModuleList(
            [
                MTLRSBlock(
                    reconstruction_module_params=reconstruction_module_params,
                    segmentation_module_params=segmentation_module_params,
                    input_channels=cfg_dict.get("segmentation_module_input_channels", 2),
                    magnitude_input=cfg_dict.get("magnitude_input", False),
                    fft_centered=cfg_dict.get("fft_centered", False),
                    fft_normalization=cfg_dict.get("fft_normalization", "backward"),
                    spatial_dims=cfg_dict.get("spatial_dims", (-2, -1)),
                    coil_dim=self.coil_dim,
                    dimensionality=cfg_dict.get("dimensionality", 2),
                    consecutive_slices=self.consecutive_slices,
                    coil_combination_method=cfg_dict.get("coil_combination_method", "SENSE"),
                    normalize_segmentation_output=cfg_dict.get("normalize_segmentation_output", True),
                )
                for _ in range(self.rs_cascades)
            ]
        )

        self.task_adaption_type = cfg_dict.get("task_adaption_type", "multi_task_learning")

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        init_reconstruction_pred: torch.Tensor,
        target_reconstruction: torch.Tensor,
        hx: torch.Tensor = None,
        sigma: float = 1.0,
    ) -> Tuple[Union[List, torch.Tensor], torch.Tensor]:
        """Forward pass of :class:`MTLRS`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        init_reconstruction_pred : torch.Tensor
            Initial reconstruction prediction. Shape [batch_size, n_x, n_y, 2]
        target_reconstruction : torch.Tensor
            Target reconstruction. Shape [batch_size, n_x, n_y, 2]
        hx : torch.Tensor, optional
            Initial hidden state for the RNN. Default is ``None``.
        sigma : float, optional
            Standard deviation of the noise. Default is ``1.0``.

        Returns
        -------
        Tuple[Union[List, torch.Tensor], torch.Tensor]
            Tuple containing the predicted reconstruction and segmentation.
        """
        pred_reconstructions = []
        for cascade in self.rs_module:
            pred_reconstruction, pred_segmentation, hx = cascade(
                y=y,
                sensitivity_maps=sensitivity_maps,
                mask=mask,
                init_reconstruction_pred=init_reconstruction_pred,
                target_reconstruction=target_reconstruction,
                hx=hx,
                sigma=sigma,
            )
            pred_reconstructions.append(pred_reconstruction)
            init_reconstruction_pred = pred_reconstruction[-1][-1]

            if self.task_adaption_type == "multi_task_learning":
                hidden_states = [
                    torch.cat(
                        [torch.abs(init_reconstruction_pred.unsqueeze(self.coil_dim) * pred_segmentation)]
                        * (f // self.segmentation_module_output_channels),
                        dim=self.coil_dim,
                    )
                    for f in self.reconstruction_module_recurrent_filters
                    if f != 0
                ]

                if self.consecutive_slices > 1:
                    hx = [x.unsqueeze(1) for x in hx]

                # Check if the concatenated hidden states are the same size as the hidden state of the RNN
                if hidden_states[0].shape[self.coil_dim] != hx[0].shape[self.coil_dim]:
                    prev_hidden_states = hidden_states
                    hidden_states = []
                    for hs in prev_hidden_states:
                        new_hidden_state = hs
                        for _ in range(hx[0].shape[1] - prev_hidden_states[0].shape[1]):
                            new_hidden_state = torch.cat(
                                [new_hidden_state, torch.zeros_like(hx[0][:, 0, :, :]).unsqueeze(self.coil_dim)],
                                dim=self.coil_dim,
                            )
                        hidden_states.append(new_hidden_state)

                hx = [hx[i] + hidden_states[i] for i in range(len(hx))]

            init_reconstruction_pred = torch.view_as_real(init_reconstruction_pred)

        return pred_reconstructions, pred_segmentation

    def process_reconstruction_loss(  # noqa: MC0001
        self,
        target: torch.Tensor,
        prediction: Union[List[List[torch.Tensor]], List[torch.Tensor], torch.Tensor],
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        attrs: Union[Dict, torch.Tensor],
        r: Union[int, torch.Tensor],
        loss_func: torch.nn.Module,
    ) -> torch.Tensor:
        """Processes the reconstruction loss for the CIRIM model. It differs from the base class in that it can handle
        multiple cascades and time steps.

        Parameters
        ----------
        target : torch.Tensor
            Target data of shape [batch_size, n_x, n_y, 2].
        prediction : Union[list, torch.Tensor]
            Prediction(s) of shape [batch_size, n_x, n_y, 2].
        sensitivity_maps : torch.Tensor
            Sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2]. It will be used if self.ssdu is True, to
            expand the target and prediction to multiple coils.
        mask : torch.Tensor
            Mask of shape [batch_size, n_x, n_y, 2]. It will be used if self.ssdu is True, to enforce data consistency
            on the prediction.
        attrs : Dict
            Attributes of the data with pre normalization values.
        r : int
            The selected acceleration factor.
        loss_func : torch.nn.Module
            Loss function. Must be one of {torch.nn.L1Loss(), torch.nn.MSELoss(),
            atommic.collections.reconstruction.losses.ssim.SSIMLoss()}. Default is ``torch.nn.L1Loss()``.

        Returns
        -------
        loss: torch.FloatTensor
            If self.accumulate_loss is True, returns an accumulative result of all intermediate losses.
            Otherwise, returns the loss of the last intermediate loss.
        """
        # If kspace reconstruction loss is used, the target needs to be transformed to k-space.
        if self.kspace_reconstruction_loss:
            # If inputs are complex, then they need to be viewed as real.
            if target.shape[-1] != 2 and torch.is_complex(target):
                target = torch.view_as_real(target)
            # If SSDU is used, then the coil-combined inputs need to be expanded to multiple coils using the
            # sensitivity maps.
            if self.ssdu:
                target = expand_op(target, sensitivity_maps, self.coil_dim)
            # Transform to k-space.
            target = fft2(target, self.fft_centered, self.fft_normalization, self.spatial_dims)
            # Ensure loss inputs are both viewed in the same way.
            target = self.__abs_output__(target / torch.max(torch.abs(target)))
        elif not self.unnormalize_loss_inputs:
            target = self.__abs_output__(target / torch.max(torch.abs(target)))

        def compute_reconstruction_loss(t, p, s):
            if self.unnormalize_loss_inputs:
                # we do the unnormalization here to avoid explicitly iterating through list of predictions, which
                # might be a list of lists.
                t, p, s = self.__unnormalize_for_loss_or_log__(t, p, s, attrs, r)

            # If kspace reconstruction loss is used, the target needs to be transformed to k-space.
            if self.kspace_reconstruction_loss:
                # If inputs are complex, then they need to be viewed as real.
                if p.shape[-1] != 2 and torch.is_complex(p):
                    p = torch.view_as_real(p)
                # If SSDU is used, then the coil-combined inputs need to be expanded to multiple coils using the
                # sensitivity maps.
                if self.ssdu:
                    p = expand_op(p, s, self.coil_dim)
                # Transform to k-space.
                p = fft2(p, self.fft_centered, self.fft_normalization, self.spatial_dims)
                # If SSDU is used, then apply the mask to the prediction to enforce data consistency.
                if self.ssdu:
                    p = p * mask
                # Ensure loss inputs are both viewed in the same way.
                p = self.__abs_output__(p / torch.max(torch.abs(p)))
            elif not self.unnormalize_loss_inputs:
                p = self.__abs_output__(p / torch.max(torch.abs(p)))

            if "ssim" in str(loss_func).lower():
                p = torch.abs(p / torch.max(torch.abs(p)))
                t = torch.abs(t / torch.max(torch.abs(t)))

                return loss_func(
                    t,
                    p,
                    data_range=torch.tensor([max(torch.max(t).item(), torch.max(p).item())]).unsqueeze(dim=0).to(t),
                )

            return loss_func(t, p)

        if self.accumulate_predictions:
            rs_cascades_weights = torch.logspace(-1, 0, steps=len(prediction)).to(target.device)
            rs_cascades_loss = []
            for rs_cascade_pred in prediction:
                cascades_weights = torch.logspace(-1, 0, steps=len(rs_cascade_pred)).to(target.device)
                cascades_loss = []
                for cascade_pred in rs_cascade_pred:
                    time_steps_weights = torch.logspace(-1, 0, steps=self.time_steps).to(target.device)
                    time_steps_loss = [
                        compute_reconstruction_loss(target, time_step_pred, sensitivity_maps)
                        for time_step_pred in cascade_pred
                    ]
                    cascade_loss = sum(x * w for x, w in zip(time_steps_loss, time_steps_weights)) / self.time_steps
                    cascades_loss.append(cascade_loss)
                rs_cascade_loss = sum(x * w for x, w in zip(cascades_loss, cascades_weights)) / len(rs_cascade_pred)
                rs_cascades_loss.append(rs_cascade_loss)
            loss = sum(x * w for x, w in zip(rs_cascades_loss, rs_cascades_weights)) / len(prediction)
        else:
            # keep the last prediction of the last cascade of the last rs cascade
            prediction = prediction[-1][-1][-1]
            loss = compute_reconstruction_loss(target, prediction, sensitivity_maps)
        return loss



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/classes/common.py

import inspect
import traceback
from abc import ABC
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from functools import total_ordering
from typing import Dict, Iterable, List, Optional, Tuple, Union

import hydra
import torch
import wrapt
from huggingface_hub import HfApi, HfFolder, ModelFilter, hf_hub_download
from huggingface_hub.hf_api import ModelInfo
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic import __version__
from atommic.core.connectors.save_restore_connector import SaveRestoreConnector
from atommic.core.neural_types.comparison import NeuralTypeComparisonResult
from atommic.core.neural_types.neural_type import NeuralType
from atommic.utils import logging, model_utils

__all__ = ["Typing", "FileIO", "Model", "PretrainedModelInfo", "Serialization", "is_typecheck_enabled", "typecheck"]

_TYPECHECK_ENABLED = True
_HAS_HYDRA = True


def is_typecheck_enabled():
    """Getter method for typechecking state."""
    return _TYPECHECK_ENABLED


@dataclass
class TypecheckMetadata:
    """Metadata class for input/output neural types.

    Parameters
    ----------
    original_types : Dict[str, NeuralType]
        Preserve the dictionary of type information provided.
    ignore_collections : bool
        For backward compatibility, container support can be disabled explicitly using this flag. When set to True, all
         nesting is ignored and nest-depth checks are skipped.
    mandatory_types : Dict[str, NeuralType]
        Sub-dictionary of `original_types` which contains only those types which are mandatory to include when calling
        the function.
    base_types : Dict[str, NeuralType]
        Dictionary of flattened `str: NeuralType` definitions, disregarding the nest level details into appropriate
        arguments.
    container_depth : Dict[str, int]
        Dictionary mapping `str: int` - such that the valid depth of the nest of this neural type is recorded.
    has_container_types : bool
        Bool flag declaring if any of the neural types declares a container nest in its signature.
    is_singular_container_type : bool
        Bool flag declaring if this is a single Neural Type with a container nest in its signature. Required for
        supporting python list expansion in return statement.
    """

    original_types: Dict[str, NeuralType]
    ignore_collections: bool

    mandatory_types: Dict[str, NeuralType] = field(init=False)
    base_types: Dict[str, NeuralType] = field(init=False)

    container_depth: Dict[str, int] = field(init=False)
    has_container_types: bool = field(init=False)
    is_singular_container_type: bool = field(init=False)

    def __post_init__(self):
        """Post init function to compute metadata."""
        has_container_types = any(isinstance(type_val, (list, tuple)) for type_val in self.original_types.values())

        self.has_container_types = has_container_types

        # If only one NeuralType is declared, and it declares a container nest, set to True
        self.is_singular_container_type = self.has_container_types and len(self.original_types) == 1

        # If container nests are declared, flatten the nest into `base_types`
        # Also compute the nest depth for each of the NeuralTypes
        if self.has_container_types:
            self.base_types = {}
            self.container_depth = {}

            for type_key, type_val in self.original_types.items():
                depth = 0
                while isinstance(type_val, (list, tuple)):
                    if len(type_val) > 1:
                        raise TypeError(
                            f"Neural Type `{type_key}`: {type_val} definition contains more than one element when"
                            "declaring the nested container structure.\n"
                            "Please ensure that you have only 1 NeuralType inside of the entire nested structure "
                            "definition."
                        )

                    type_val = type_val[0]
                    depth += 1

                self.base_types[type_key] = type_val
                self.container_depth[type_key] = depth
        else:
            # Otherwise, simply preserve the original_types and set depth of nest to 0.
            self.base_types = self.original_types
            self.container_depth = {type_key: 0 for type_key in self.base_types.keys()}

        # Compute subset of original_types which are mandatory in the call argspec
        self.mandatory_types = {
            type_key: type_val for type_key, type_val in self.base_types.items() if not type_val.optional
        }


class Typing(ABC):
    """An interface which endows module with neural types."""

    @property
    def input_types(self) -> Optional[Dict[str, NeuralType]]:
        """Define these to enable input neural type checks."""
        return None

    @property
    def output_types(self) -> Optional[Dict[str, NeuralType]]:
        """Define these to enable output neural type checks."""
        return None

    def _validate_input_types(self, input_types=None, ignore_collections=False, **kwargs):
        """This function does a few things.

            1) It ensures that len(self.input_types <non-optional>) <= len(kwargs) <= len(self.input_types).

            2) For each (keyword name, keyword value) passed as input to the wrapped function:

                - Check if the keyword name exists in the list of valid self.input_types names.

                - Check if keyword value has the `neural_type` property.

                    - If it does, then perform a comparative check and assert that neural types are compatible (SAME
                        or GREATER).

                - Check if keyword value is a container type (list or tuple). If yes, then perform the elementwise
                    test of neural type above on each element of the nested structure, recursively.

        Parameters
        ----------
        input_types : class
            Either the `input_types` defined at class level, or the local function overridden type definition.
        ignore_collections : bool
            For backward compatibility, container support can be disabled explicitly using this flag. When set to True,
             all nesting is ignored and nest-depth checks are skipped.
        kwargs : Dict[str, Any]
            Dictionary of argument_name:argument_value pairs passed to the wrapped function upon call.
        """
        # TODO: Properly implement this
        if input_types is None:
            return
        # Precompute metadata
        metadata = TypecheckMetadata(original_types=input_types, ignore_collections=ignore_collections)

        total_input_types = len(input_types)
        mandatory_input_types = len(metadata.mandatory_types)

        # Allow number of input arguments to be <= total input neural types.
        if len(kwargs) < mandatory_input_types or len(kwargs) > total_input_types:
            raise TypeError(
                f"Number of input arguments provided ({len(kwargs)}) is not as expected. Function has "
                f"{total_input_types} total inputs with {mandatory_input_types} mandatory inputs."
            )

        for key, value in kwargs.items():
            # Check if keys exists in the defined input types
            if key not in input_types:
                raise TypeError(
                    f"Input argument {key} has no corresponding input_type match. "
                    f"Existing input_types = {input_types.keys()}"
                )

                # Perform neural type check
            if hasattr(value, "neural_type") and metadata.base_types[key].compare(value.neural_type) not in (
                NeuralTypeComparisonResult.SAME,
                NeuralTypeComparisonResult.GREATER,
            ):
                error_msg = [
                    f"{input_types[key].compare(value.neural_type)} :",
                    f"Input type expected : {input_types[key]}",
                    f"Input type found : {value.neural_type}",
                    f"Argument: {key}",
                ]
                for i, dict_tuple in enumerate(metadata.base_types[key].elements_type.type_parameters.items()):
                    error_msg.insert(i + 2, f"  input param_{i} : {dict_tuple[0]}: {dict_tuple[1]}")
                error_msg.extend(
                    f"  input param_{i} : {dict_tuple[0]}: {dict_tuple[1]}"
                    for i, dict_tuple in enumerate(value.neural_type.elements_type.type_parameters.items())
                )

                raise TypeError("\n".join(error_msg))

                # Perform input n dim check
            if hasattr(value, "shape"):
                value_shape = value.shape
                type_shape = metadata.base_types[key].axes
                if type_shape is not None and len(value_shape) != len(tuple(type_shape)):
                    name = key

                    raise TypeError(
                        f"Input shape mismatch occurred for {name} in module {self.__class__.__name__} : \n"
                        f"Input shape expected = {metadata.base_types[name].axes} | \n"
                        f"Input shape found : {value_shape}"
                    )

            elif isinstance(value, (list, tuple)):
                for val in value:
                    # This initiates a DFS, tracking the depth count as it goes along the nested structure.
                    # Initial depth is 1 as we consider the current loop to be the 1st step inside the nest.
                    self.__check_neural_type(val, metadata, depth=1, name=key)

    def _attach_and_validate_output_types(self, out_objects, ignore_collections=False, output_types=None):
        """This function does a few things.
            1) It ensures that len(out_object) == len(self.output_types).
            2) If the output is a tensor (or list/tuple of list/tuple ... of tensors), it
                attaches a neural_type to it. For objects without the neural_type attribute,
                such as python objects (dictionaries and lists, primitive data types, structs),
                no neural_type is attached.
                Note: tensor.neural_type is only checked during _validate_input_types which is
                called prior to forward().

        Parameters
        ----------
        output_types : class
            Either the `output_types` defined at class level, or the local function overridden type definition.
        ignore_collections : bool
            For backward compatibility, container support can be disabled explicitly using this flag. When set to True,
             all nesting is ignored and nest-depth checks are skipped.
        out_objects : Dict[str, Any]
            The outputs of the wrapped function.
        """
        # TODO: Properly implement this
        if output_types is None:
            return
        # Precompute metadata
        metadata = TypecheckMetadata(original_types=output_types, ignore_collections=ignore_collections)
        out_types_list = list(metadata.base_types.items())
        mandatory_out_types_list = list(metadata.mandatory_types.items())

        # First convert all outputs to list/tuple format to check correct number of outputs
        if isinstance(out_objects, (list, tuple)):
            out_container = out_objects  # can be any rank nested structure
        else:
            out_container = [out_objects]

        # If this neural type has a *single output*, with *support for nested outputs*,
        # then *do not* perform any check on the number of output items against the number
        # of neural types (in this case, 1).
        # This is done as python will *not* wrap a single returned list into a tuple of length 1,
        # instead opting to keep the list intact. Therefore len(out_container) in such a case
        # is the length of all the elements of that list - each of which has the same corresponding
        # neural type (defined as the singular container type).
        if metadata.is_singular_container_type:
            pass

        elif len(out_container) > len(out_types_list) or len(out_container) < len(mandatory_out_types_list):
            raise TypeError(
                f"Number of output arguments provided ({len(out_container)}) is not as expected. It should be larger "
                f"than {len(out_types_list)} and less than {len(mandatory_out_types_list)}.\nThis can be either "
                "because insufficient/extra number of output NeuralTypes were provided,or the provided NeuralTypes "
                f"{output_types} should enable container support (add '[]' to the NeuralType definition)"
            )

            # Attach types recursively, if possible
        if not isinstance(out_objects, tuple) and not isinstance(out_objects, list):
            # Here, out_objects is a single object which can potentially be attached with a NeuralType
            try:
                out_objects.neural_type = out_types_list[0][1]
            except AttributeError:
                pass

                # Perform output n dim check
            if hasattr(out_objects, "shape"):
                value_shape = out_objects.shape
                type_shape = out_types_list[0][1].axes
                if type_shape is not None and len(value_shape) != len(type_shape):
                    name = out_types_list[0][0]

                    raise TypeError(
                        f"Output shape mismatch occurred for {name} in module {self.__class__.__name__} : \n"
                        f"Output shape expected = {type_shape} | \n"
                        f"Output shape found : {value_shape}"
                    )

        elif metadata.is_singular_container_type:
            depth = 0 if len(out_objects) == 1 and isinstance(out_objects, tuple) else 1
            for res in out_objects:
                self.__attach_neural_type(res, metadata, depth=depth, name=out_types_list[0][0])
        else:
            # If more than one item is returned in a return statement, python will wrap the output with an outer tuple.
            # Therefore, there must be a 1:1 correspondence of the output_neural type (with or without nested
            # structure) to the actual output (whether it is a single object or a nested structure of objects).
            # Therefore, in such a case, we "start" the DFS at depth 0 - since the recursion is being applied on 1
            # neural type : 1 output struct (single or nested output). Since we are guaranteed that the outer tuple
            # will be built by python, assuming initial depth of 0 is appropriate.
            for ind, res in enumerate(out_objects):
                self.__attach_neural_type(res, metadata, depth=0, name=out_types_list[ind][0])

    def __check_neural_type(self, obj, metadata, depth: int, name: str = None):
        """Checks if the object is of the correct type, and attaches the correct NeuralType.

        Parameters
        ----------
        obj : object
            Any python object that can be assigned to a value.
        metadata : object
            TypecheckMetadata object.
        depth : int
            Current depth of the recursion.
        name : str
            Optional name used of the source object, when an error is raised.
        """
        if isinstance(obj, (tuple, list)):
            for elem in obj:
                self.__check_neural_type(elem, metadata, depth + 1, name=name)
            return  # after processing nest, return to avoid testing nest itself

        type_val = metadata.base_types[name]

        # If nest depth doesnt match neural type structure depth, raise an error
        if not metadata.ignore_collections and depth != metadata.container_depth[name]:
            raise TypeError(
                "While checking input neural types,\n"
                "Nested depth of value did not match container specification:\n"
                f"Current nested depth of NeuralType '{name}' ({type_val}): {depth}\n"
                f"Expected nested depth : {metadata.container_depth[name]}"
            )

        if hasattr(obj, "neural_type") and type_val.compare(obj.neural_type) not in (
            NeuralTypeComparisonResult.SAME,
            NeuralTypeComparisonResult.GREATER,
        ):
            raise TypeError(
                f"{type_val.compare(obj.neural_type)} : \n"
                f"Input type expected = {type_val} | \n"
                f"Input type found : {obj.neural_type}"
            )

        # Perform input n dim check
        if hasattr(obj, "shape"):
            value_shape = obj.shape
            type_shape = type_val.axes

            if type_shape is not None and len(value_shape) != len(type_shape):
                raise TypeError(
                    f"Input shape mismatch occurred for {name} in module {self.__class__.__name__} : \n"
                    f"Input shape expected = {type_shape} | \n"
                    f"Input shape found : {value_shape}"
                )

    def __attach_neural_type(self, obj, metadata, depth: int, name: str = None):
        """Attach NeuralType to the object.

        Parameters
        ----------
        obj : object
            Any python object that can be assigned to a value.
        metadata : object
            TypecheckMetadata object.
        depth : int
            Current depth of the recursion.
        name : str
            Optional name used of the source object, when an error is raised.
        """
        if isinstance(obj, (tuple, list)):
            for elem in obj:
                self.__attach_neural_type(elem, metadata, depth=depth + 1, name=name)
            return  # after processing nest, return to avoid argument insertion into nest itself

        type_val = metadata.base_types[name]

        # If nest depth doesnt match neural type structure depth, raise an error
        if not metadata.ignore_collections and depth != metadata.container_depth[name]:
            raise TypeError(
                "While attaching output neural types,\n"
                "Nested depth of value did not match container specification:\n"
                f"Current nested depth of NeuralType '{name}' ({type_val}): {depth}\n"
                f"Expected nested depth : {metadata.container_depth[name]}"
            )

        try:
            obj.neural_type = type_val
        except AttributeError:
            pass

        # Perform output n dim check
        if hasattr(obj, "shape"):
            value_shape = obj.shape
            type_shape = type_val.axes

            if type_shape is not None and len(value_shape) != len(type_shape):
                raise TypeError(
                    f"Output shape mismatch occurred for {name} in module {self.__class__.__name__} : \n"
                    f"Output shape expected = {type_shape} | \n"
                    f"Output shape found : {value_shape}"
                )


class Serialization(ABC):
    """Base class for serialization."""

    @classmethod
    def from_config_dict(cls, config: "DictConfig", trainer: Optional[Trainer] = None):  # noqa: MC0001
        """Instantiates object using DictConfig-based configuration"""
        # Resolve the config dict
        if _HAS_HYDRA:
            if isinstance(config, DictConfig):
                config = OmegaConf.to_container(config, resolve=True)
                config = OmegaConf.create(config)
                OmegaConf.set_struct(config, True)

            config = model_utils.maybe_update_config_version(config)

        # Hydra 0.x API
        if ("cls" in config or "target" in config) and "params" in config and _HAS_HYDRA:
            # regular hydra-based instantiation
            instance = hydra.utils.instantiate(config=config)
        elif "_target_" in config and _HAS_HYDRA:
            # regular hydra-based instantiation
            instance = hydra.utils.instantiate(config=config)
        else:
            instance = None
            prev_error = ""

            # Attempt class path resolution from config `target` class (if it exists)
            if "target" in config:
                # No guarantee that this is a omegaconf class
                target_cls = config["target"]
                imported_cls = None
                try:
                    # try to import the target class
                    imported_cls = model_utils.import_class_by_path(target_cls)
                    # use subclass instead
                    if issubclass(cls, imported_cls):
                        imported_cls = cls
                    if (
                        accepts_trainer :=  # pylint: disable=unused-variable
                        # check if the target class accepts a trainer argument
                        Serialization._inspect_signature_for_trainer(imported_cls)
                    ):
                        # Create a dummy PL trainer object
                        instance = imported_cls(cfg=config, trainer=trainer)
                    else:
                        instance = imported_cls(cfg=config)

                except Exception as e:
                    tb = traceback.format_exc()
                    prev_error = f"Model instantiation failed.\nTarget class: {target_cls}\nError: {e}\n{tb}"
                    logging.debug(prev_error + "\n falling back to 'cls'.")
            # target class resolution was unsuccessful, fall back to current `cls`
            if instance is None:
                try:
                    if accepts_trainer := Serialization._inspect_signature_for_trainer(cls):  # noqa: F841
                        instance = cls(cfg=config, trainer=trainer)  # type: ignore
                    else:
                        instance = cls(cfg=config)  # type: ignore
                except Exception as e:
                    # report saved errors, if any, and raise the current error
                    if prev_error:
                        logging.error(f"{prev_error}")
                    raise e from e

        if not hasattr(instance, "_cfg"):
            instance._cfg = config  # pylint: disable=protected-access
        return instance

    def to_config_dict(self) -> "DictConfig":
        """Returns object's configuration to config dictionary"""
        if (hasattr(self, "_cfg")) and (
            self._cfg is not None  # type: ignore  # pylint: disable=access-member-before-definition
        ):
            # Resolve the config dict
            if (_HAS_HYDRA) and (
                isinstance(self._cfg, DictConfig)  # type: ignore  # pylint: disable=access-member-before-definition
            ):
                config = OmegaConf.to_container(
                    self._cfg, resolve=True  # type: ignore  # pylint: disable=access-member-before-definition
                )
                config = OmegaConf.create(config)
                OmegaConf.set_struct(config, True)

                config = model_utils.maybe_update_config_version(config)

            self._cfg = config

            return self._cfg
        raise NotImplementedError(
            "to_config_dict() can currently only return object._cfg but current object does not have it."
        )

    @classmethod
    def _inspect_signature_for_trainer(cls, check_cls):
        """Inspects the signature of the class to see if it accepts a trainer argument."""
        if hasattr(check_cls, "__init__"):
            signature = inspect.signature(check_cls.__init__)
            if "trainer" in signature.parameters:
                return True
        return False


class FileIO(ABC):
    """Base class for file IO."""

    def save_to(self, save_path: str):
        """Standardized method to save a tarfile containing the checkpoint, config, and any additional artifacts.
        Implemented via :meth:`atommic.core.connectors.save_restore_connector.SaveRestoreConnector.save_to`.

        Parameters
        ----------
        save_path : str
            Path to save the checkpoint to.
        """
        raise NotImplementedError()

    @classmethod
    def restore_from(
        cls,
        restore_path: str,
        override_config_path: Optional[str] = None,
        map_location: Optional[torch.device] = None,
        strict: bool = True,
        return_config: bool = False,
        trainer: Optional[Trainer] = None,
        save_restore_connector: SaveRestoreConnector = None,
    ):
        """Restores model instance (weights and configuration) from a .atommic file.

        Parameters
        ----------
        restore_path : str
            Path to .atommic file from which model should be instantiated.
        override_config_path : str, optional
            Path to .yaml file containing the configuration to override the one in the .atommic file.
        map_location : torch.device, optional
            Device to map the instantiated model to. Default  is ``None``, it will select a GPU if available, falling
            back to CPU otherwise.
        strict : bool, optional
            Passed to load_state_dict. Default is ``True``.
        return_config : bool, optional
            If True, returns the underlying config of the restored model as an OmegaConf DictConfig object without
            instantiating the model.
        trainer : Trainer, optional
            If provided, will be used to instantiate the model.
        save_restore_connector : SaveRestoreConnector, optional
            An optional SaveRestoreConnector object that defines the implementation of the restore_from() method.
        """
        raise NotImplementedError()

    @classmethod
    def from_config_file(cls, path2yaml_file: str):
        """Instantiates an instance of atommic Model from YAML config file. Weights will be initialized randomly.

        Parameters
        ----------
        path2yaml_file : str
            Path to yaml file with model configuration.

        Returns
        -------
        atommic Model instance.
        """
        if issubclass(cls, Serialization):
            conf = OmegaConf.load(path2yaml_file)
            return cls.from_config_dict(config=conf)
        raise NotImplementedError()

    def to_config_file(self, path2yaml_file: str):
        """Saves current instance's configuration to YAML config file. Weights will not be saved.

        Parameters
        ----------
        path2yaml_file : str
            Path to yaml file with model configuration.
        """
        if hasattr(self, "_cfg"):
            self._cfg = model_utils.maybe_update_config_version(self._cfg)  # type: ignore
            with open(path2yaml_file, "w", encoding="utf-8") as fout:
                OmegaConf.save(config=self._cfg, f=fout, resolve=True)
        else:
            raise NotImplementedError()


@total_ordering
@dataclass
class PretrainedModelInfo:
    """Class to store information about a pretrained model."""

    pretrained_model_name: str
    description: str
    location: str
    class_: Union["Model", None] = None
    aliases: Union[List[str], None] = None

    def __repr__(self):
        """Return a string representation of the object."""
        base = self.__class__.__name__
        extras = (
            "pretrained_model_name={pretrained_model_name},\n\t"
            "description={description},\n\t"
            "location={location}".format(**self.__dict__)
        )

        if self.class_ is not None:
            extras = "{extras},\n\t" "class_={class_}".format(extras=extras, **self.__dict__)

        return f"{base}(\n\t{extras}\n)"

    def __hash__(self):
        """Return a hash of the object."""
        return hash(self.location)

    def __eq__(self, other):
        """Return True if self is equal to other."""
        # another object is equal to self, if it's hash is equal to hash(self)
        return hash(self) == hash(other) or self.pretrained_model_name == other.pretrained_model_name

    def __lt__(self, other):
        """Return True if self is less than other."""
        return self.pretrained_model_name < other.pretrained_model_name


class Model(Typing, Serialization, FileIO, ABC):  # type: ignore
    """Abstract class offering interface which should be implemented by all atommic models."""

    @classmethod
    def list_available_models(cls) -> Optional[List[PretrainedModelInfo]]:
        """Should list all pre-trained models available.
        Note: There is no check that requires model names and aliases to be unique. In the case of a collision,
        whatever model (or alias) is listed first in the returned list will be instantiated.

        Returns
        -------
        A list of PretrainedModelInfo entries.
        """
        raise NotImplementedError()

    @classmethod
    def search_huggingface_models(
        cls, model_filter: Optional[Union[ModelFilter, List[ModelFilter]]] = None
    ) -> List[ModelInfo]:
        """Should list all pre-trained models available via Hugging Face Hub.

        The following metadata can be passed via the `model_filter` for additional results.

        .. metadata::
            resolve_card_info: Bool flag, if set, returns the model card metadata. Default: False.
            limit_results: Optional int, limits the number of results returned.

        .. code-block:: python

            # You can replace <DomainSubclass> with any subclass of ModelPT.
            from atommic.core import ModelPT

            # Get default ModelFilter
            filt = <DomainSubclass>.get_hf_model_filter()

            # Make any modifications to the filter as necessary
            filt.language = [...]
            filt.task = ...
            filt.tags = [...]

            # Add any metadata to the filter as needed
            filt.limit_results = 5

            # Obtain model info
            model_infos = <DomainSubclass>.search_huggingface_models(model_filter=filt)

            # Browse through cards and select an appropriate one
            card = model_infos[0]

            # Restore model using `modelId` of the card.
            model = ModelPT.from_pretrained(card.modelId)

        Parameters
        ----------
        model_filter : Optional ModelFilter or List[ModelFilter] (from Hugging Face Hub)
            Filters the returned list of compatible model cards, and selects all results from each filter.
            Users can then use `model_card.modelId` in `from_pretrained()` to restore a atommic Model.
            If no ModelFilter is provided, uses the classes default filter as defined by `get_hf_model_filter()`.

        Returns
        -------
        list
            A list of ModelInfo entries.
        """
        # Resolve model filter if not provided as argument
        if model_filter is None:
            model_filter = cls.get_hf_model_filter()

        # If single model filter, wrap into list
        if not isinstance(model_filter, Iterable):
            model_filter = [model_filter]

        # Inject `atommic` library filter
        for mfilter in model_filter:
            if isinstance(mfilter.library, str) and mfilter.library != "atommic":
                logging.warning(
                    f"Model filter's `library` tag updated be `atommic`. Original value: {mfilter.library}"
                )
                mfilter.library = "atommic"

            elif isinstance(mfilter, Iterable) and "atommic" not in mfilter.library:  # type: ignore
                logging.warning(
                    "Model filter's `library` list updated to include `atommic`. "
                    f"Original value: {mfilter.library}"  # type: ignore
                )
                mfilter.library = list(mfilter)  # type: ignore
                mfilter.library.append("atommic")  # type: ignore

        # Check if api token exists, use if it does
        is_token_available = HfFolder.get_token() is not None

        # Search for all valid models after filtering
        api = HfApi()

        # Setup extra arguments for model filtering
        all_results = []  # type: List[ModelInfo]

        for mfilter in model_filter:
            cardData = None
            limit = None

            if hasattr(mfilter, "resolve_card_info") and mfilter.resolve_card_info is True:
                cardData = True

            if hasattr(mfilter, "limit_results") and mfilter.limit_results is not None:
                limit = mfilter.limit_results

            results = api.list_models(  # pylint: disable=unexpected-keyword-arg
                filter=mfilter,
                use_auth_token=is_token_available,
                sort="lastModified",
                direction=-1,
                cardData=cardData,
                limit=limit,
            )  # type: List[ModelInfo]

            all_results.extend(results)

        return all_results

    @classmethod
    def get_available_model_names(cls) -> List[str]:
        """Returns the list of model names available. To get the complete model description use
        list_available_models().

        Returns
        -------
        A list of model names.
        """
        return (
            [model.pretrained_model_name for model in cls.list_available_models()]  # type: ignore
            if cls.list_available_models() is not None
            else []
        )

    @classmethod
    def get_hf_model_filter(cls) -> ModelFilter:
        """Generates a filter for HuggingFace models.

        Additionally, includes default values of some metadata about results returned by the Hub.

        .. metadata::
            resolve_card_info: Bool flag, if set, returns the model card metadata. Default: False.
            limit_results: Optional int, limits the number of results returned.

        Returns
        -------
        list
            A Hugging Face Hub ModelFilter object.
        """
        model_filter = ModelFilter(library="atommic")

        # Attach some additional info
        model_filter.resolve_card_info = False
        model_filter.limit_results = None

        return model_filter

    @classmethod
    def from_pretrained(
        cls,
        model_name: str,
        refresh_cache: bool = False,
        override_config_path: Optional[str] = None,
        map_location: Optional[torch.device] = None,
        strict: bool = True,
        return_config: bool = False,
        trainer: Optional[Trainer] = None,
        save_restore_connector: SaveRestoreConnector = None,
        return_only_atommic_model_file_in_cache: bool = False,
    ):
        """Instantiates an instance of atommic. Use restore_from() to instantiate from a local .atommic file.

        Parameters
        ----------
        model_name : str
            The name of the model to instantiate.
        refresh_cache : bool, optional
            If set to True, then when fetching from cloud, this will re-fetch the file from cloud even if it is already
            found in a cache locally.
        override_config_path : str, optional
            Path to a yaml config that will override the internal config file.
        map_location : torch.device, optional
            Optional torch.device() to map the instantiated model to a device. Default is ``None``. It will select a
            GPU if available, falling back to CPU otherwise.
        strict : bool, optional
            Passed to torch.load_state_dict. Default is ``True``.
        return_config : bool, optional
            If set to true, will return just the underlying config of the restored model as an OmegaConf/DictConfig
            object without instantiating the model.
        trainer : Trainer, optional
            Optional Trainer objects to use for restoring the model.
        save_restore_connector : SaveRestoreConnector, optional
            Optional SaveRestoreConnector object to use for restoring the model.
        return_only_atommic_model_file_in_cache : bool, optional
            If set to True, will return the path to the atommic model file in the cache, without instantiating the
            model.

        Returns
        -------
            A model instance of a particular model class or its underlying config (if return_config is set).
        """
        if save_restore_connector is None:
            save_restore_connector = SaveRestoreConnector()

        # Resolve if the pretrained model name is on HF Hub source
        (
            class_,
            atommic_model_file_in_cache,
        ) = cls._get_hf_hub_pretrained_model_info(  # type: ignore
            cls, model_name=model_name, refresh_cache=refresh_cache  # type: ignore
        )
        if return_only_atommic_model_file_in_cache:
            return atommic_model_file_in_cache

        instance, state_dict = class_.restore_from(  # type: ignore
            restore_path=atommic_model_file_in_cache,
            override_config_path=override_config_path,
            map_location=map_location,
            strict=strict,
            return_config=return_config,
            trainer=trainer,
            save_restore_connector=save_restore_connector,
        )
        return instance, state_dict

    def _get_hf_hub_pretrained_model_info(cls, model_name: str, refresh_cache: bool = False) -> Tuple[type, str]:
        """Resolve the HuggingFace Hub model pretrained information given a model name.

        The model name must be of general syntax ``{source_repo}/{model_name}``.

        .. note:
            This allows public, externally contributed models to be run freely using atommic.

        Parameters
        ----------
        model_name : str
            Name of the model. Must be the original name or an alias of the model, without any '/'.
        refresh_cache : bool
            Determines whether cache must be refreshed (model is re-downloaded).

        Returns
        -------
        tuple
            A tuple of details describing :
                -   The resolved class of the model. Since the source is external to atommic, always default to using
                    the calling class. Depend on target class resolution by restore_from() for calling the correct
                    class.
                -   The path to the atommic model (.atommic file) in some cached directory (managed by HF Hub).
        """
        # Resolve the model name without origin for filename
        resolved_username = model_name.split("huggingface.co/")[-1].split("/")[0]
        resolved_model_filename = model_name.split(resolved_username)[-1].split("/blob/main/")[0].strip("/")
        resolved_stem = model_name.split("/")[-1]

        # Check if api token exists, use if it does
        is_token_available = HfFolder.get_token() is not None

        # Try to load the model from the Huggingface Hub
        path = hf_hub_download(
            repo_id=f'{resolved_username}/{resolved_model_filename}',
            filename=resolved_stem,
            library_name="atommic",
            library_version=__version__,
            force_download=refresh_cache,
            use_auth_token=is_token_available,
        )

        # Cannot pre-resolve the specific class without double instantiation (first for config, second for model
        # params). Default to current class, and perform basic class path resolution (handled via restore_from() +
        # target class)
        class_ = cls
        return class_, path  # type: ignore


class typecheck:
    """A decorator which performs input-output neural type checks, and attaches neural types to the output of the
    function that it wraps. Requires that the class inherit from `atommic.core.Typing` in order to perform type
    checking, and will raise an error if that is not the case.

    # Usage (Class level type support)
    .. code-block:: python

        @typecheck()
        def fn(self, arg1, arg2, ...):

    # Usage (Function level type support)
    .. code-block:: python

        @typecheck(input_types=..., output_types=...)
        def fn(self, arg1, arg2, ...):

    Points to be noted:
        1) The brackets () in `@typecheck()` are necessary. You will encounter a TypeError: __init__() takes 1 \
        positional argument but X were given without those brackets.
        2) The function can take any number of positional arguments during definition. When you call this function, \
        all arguments must be passed using kwargs only.
    """

    class TypeState(Enum):
        """
        Placeholder to denote the default value of type information provided.
        If the constructor of this decorator is used to override the class level type definition, this enum value
        indicate that types will be overridden.
        """

        UNINITIALIZED = 0

    def __init__(
        self,
        input_types: Union[TypeState, Optional[Dict[str, NeuralType]]] = TypeState.UNINITIALIZED,
        output_types: Union[TypeState, Optional[Dict[str, NeuralType]]] = TypeState.UNINITIALIZED,
        ignore_collections: bool = False,
    ):
        self.input_types = input_types
        self.output_types = output_types

        self.input_override = input_types != self.TypeState.UNINITIALIZED
        self.output_override = output_types != self.TypeState.UNINITIALIZED
        self.ignore_collections = ignore_collections

    @wrapt.decorator(enabled=is_typecheck_enabled)
    def __call__(self, wrapped, instance: Typing, args, kwargs):
        """Wrapper method that can be used on any function of a class that implements :class:`~atommic.core.Typing`.
        By default, it will utilize the `input_types` and `output_types` properties of the class inheriting Typing.
        Local function level overrides can be provided by supplying dictionaries as arguments to the decorator.
        """
        if instance is None:
            raise RuntimeError("Only classes which inherit atommic.core.Typing can use this decorator !")

        if not isinstance(instance, Typing):
            raise RuntimeError("Only classes which inherit atommic.core.Typing can use this decorator !")

        if hasattr(instance, "input_ports") or hasattr(instance, "output_ports"):
            raise RuntimeError(
                "Typing requires override of `input_types()` and `output_types()`, "
                "not `input_ports() and `output_ports()`"
            )

        # Preserve type information
        if self.input_types is typecheck.TypeState.UNINITIALIZED:
            self.input_types = instance.input_types

        if self.output_types is typecheck.TypeState.UNINITIALIZED:
            self.output_types = instance.output_types

        # Resolve global type or local overridden type
        input_types = self.input_types if self.input_override else instance.input_types
        if self.output_override:
            output_types = self.output_types
        else:
            output_types = instance.output_types

        # If types are not defined, skip type checks and just call the wrapped method
        if input_types is None and output_types is None:
            return wrapped(*args, **kwargs)

        # Check that all arguments are kwargs
        if input_types is not None and len(args) > 0:
            raise TypeError("All arguments must be passed by kwargs only for typed methods")

        # Perform rudimentary input checks here
        instance._validate_input_types(input_types=input_types, ignore_collections=self.ignore_collections, **kwargs)

        # Call the method - this can be forward, or any other callable method
        outputs = wrapped(*args, **kwargs)

        instance._attach_and_validate_output_types(
            output_types=output_types, ignore_collections=self.ignore_collections, out_objects=outputs
        )

        return outputs

    @staticmethod
    def set_typecheck_enabled(enabled: bool = True):
        """Set the global typecheck flag."""
        global _TYPECHECK_ENABLED
        _TYPECHECK_ENABLED = enabled

    @staticmethod
    @contextmanager
    def disable_checks():
        """Temporarily disable type checks."""
        typecheck.set_typecheck_enabled(enabled=False)
        try:
            yield
        finally:
            typecheck.set_typecheck_enabled(enabled=True)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.common.parts import fft  # noqa: F401
from atommic.collections.common.parts.transforms import (  # noqa: F401
    N2R,
    SSDU,
    Composer,
    Cropper,
    EstimateCoilSensitivityMaps,
    GeometricDecompositionCoilCompression,
    Masker,
    MRIDataTransforms,
    NoisePreWhitening,
    Normalizer,
    SNREstimator,
    ZeroFillingPadding,
)
from atommic.collections.common.parts.utils import *  # noqa: F401



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.segmentation.metrics.segmentation_metrics import (  # noqa: F401
    asd,
    binary_cross_entropy_with_logits_metric,
    dice_metric,
    f1_per_class_metric,
    hausdorff_distance_95_metric,
    hausdorff_distance_metric,
    iou_metric,
    precision_metric,
    recall_metric,
    surface_distances,
)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/optim/optimizers.py

import copy
from functools import partial
from typing import Any, Dict, Optional, Union

import hydra
import torch
from omegaconf import DictConfig, OmegaConf
from torch import optim
from torch.optim import adadelta, adagrad, adamax, rmsprop, rprop
from torch.optim.optimizer import Optimizer

from atommic.core.conf.optimizers import OptimizerParams, get_optimizer_config, register_optimizer_params
from atommic.core.optim.adafactor import Adafactor
from atommic.core.optim.lion import Lion
from atommic.core.optim.novograd import Novograd
from atommic.core.optim.radam import RAdam
from atommic.utils.model_utils import maybe_update_config_version

AVAILABLE_OPTIMIZERS = {
    "sgd": optim.SGD,
    "adam": optim.Adam,
    "adamw": optim.AdamW,
    "adadelta": adadelta.Adadelta,
    "adamax": adamax.Adamax,
    "adagrad": adagrad.Adagrad,
    "rmsprop": rmsprop.RMSprop,
    "rprop": rprop.Rprop,
    "novograd": Novograd,
    "adafactor": Adafactor,
    "radam": RAdam,
    "lion": Lion,
}

__all__ = ["AVAILABLE_OPTIMIZERS", "get_optimizer", "register_optimizer", "parse_optimizer_args"]


def parse_optimizer_args(
    optimizer_name: str, optimizer_kwargs: Union[DictConfig, Dict[str, Any]]
) -> Union[Dict[str, Any], DictConfig]:
    """Parses a list of strings, of the format "key=value" or "key2=val1,val2,..." into a dictionary of type
    {key=value, key2=[val1, val2], ...}. This dictionary is then used to instantiate the chosen Optimizer.

    Parameters
    ----------
    optimizer_name : str
        String name of the optimizer, used for auto resolution of params.
    optimizer_kwargs : Union[DictConfig, Dict[str, Any]]
        Either a list of strings in a specified format, or a dictionary. If a dictionary is provided, it is assumed the
         dictionary is the final parsed value, and simply returned. If a list of strings is provided, each item in the
         list is parsed into a new dictionary.

    Returns
    -------
    dict
        A dictionary of the parsed arguments.
    """
    kwargs: Dict[Any, Any] = {}

    if optimizer_kwargs is None:
        return kwargs

    optimizer_kwargs = copy.deepcopy(optimizer_kwargs)
    optimizer_kwargs = maybe_update_config_version(optimizer_kwargs)

    if isinstance(optimizer_kwargs, DictConfig):
        optimizer_kwargs = OmegaConf.to_container(optimizer_kwargs, resolve=True)

    # If it is a dictionary, perform stepwise resolution
    if hasattr(optimizer_kwargs, "keys"):
        # Attempt class path resolution
        if "_target_" in optimizer_kwargs:  # captures (target, _target_)
            optimizer_kwargs_config = OmegaConf.create(optimizer_kwargs)
            optimizer_instance = hydra.utils.instantiate(optimizer_kwargs_config)  # type: DictConfig
            optimizer_instance = vars(optimizer_instance)
            return optimizer_instance

        # If class path was not provided, perhaps `name` is provided for resolution
        if "name" in optimizer_kwargs:
            # If `auto` is passed as name for resolution of optimizer name,
            # then lookup optimizer name and resolve its parameter config
            if optimizer_kwargs["name"] == "auto":
                optimizer_params_name = f"{optimizer_name}_params"
                optimizer_kwargs.pop("name")
            else:
                optimizer_params_name = optimizer_kwargs.pop("name")

            # Override arguments provided in the config yaml file
            if "params" in optimizer_kwargs:
                # If optimizer kwarg overrides are wrapped in yaml `params`
                optimizer_params_override = optimizer_kwargs.get("params")
            else:
                # If the kwargs themselves are a DictConfig
                optimizer_params_override = optimizer_kwargs

            if isinstance(optimizer_params_override, DictConfig):
                optimizer_params_override = OmegaConf.to_container(optimizer_params_override, resolve=True)

            optimizer_params_cls = get_optimizer_config(optimizer_params_name, **optimizer_params_override)

            # If we are provided just a Config object, simply return the dictionary of that object
            if optimizer_params_name is None:
                optimizer_params = vars(optimizer_params_cls)
                return optimizer_params
            # If we are provided a partial class instantiation of a Config, instantiate it and retrieve its vars
            # as a dictionary.
            # instantiate the parameters object
            optimizer_params = vars(optimizer_params_cls)
            return optimizer_params

        # simply return the dictionary that was provided
        return optimizer_kwargs

    return kwargs


def register_optimizer(name: str, optimizer: Optimizer, optimizer_params: OptimizerParams):
    """Checks if the optimizer name exists in the registry, and if it doesn't, adds it. This allows custom optimizers
    to be added and called by name during instantiation.

    Parameters
    ----------
    name : str
        Name of the optimizer. Will be used as key to retrieve the optimizer.
    optimizer : Optimizer
        Optimizer class.
    optimizer_params : OptimizerParams
        The parameters as a dataclass of the optimizer.
    """
    if name in AVAILABLE_OPTIMIZERS:
        raise ValueError(f"Cannot override pre-existing optimizers. Conflicting optimizer name = {name}")

    AVAILABLE_OPTIMIZERS[name] = optimizer

    optim_name = f"{optimizer.__name__}_params"
    register_optimizer_params(name=optim_name, optimizer_params=optimizer_params)


def get_optimizer(name: str, **kwargs: Optional[Dict[str, Any]]) -> partial:
    """Convenience method to obtain an Optimizer class and partially instantiate it with optimizer kwargs.

    Parameters
    ----------
    name : str
        Name of the optimizer. Will be used as key to retrieve the optimizer.
    kwargs :  Optional[Dict[str, Any]]
        Optional kwargs of the optimizer used during instantiation.

    Returns
    -------
    partial
        A partially instantiated Optimizer.
    """
    if name not in AVAILABLE_OPTIMIZERS:
        raise ValueError(
            f"Cannot resolve optimizer '{name}'. Available optimizers are : " f"{AVAILABLE_OPTIMIZERS.keys()}"
        )
    if name == "fused_adam" and not torch.cuda.is_available():
        raise ValueError("CUDA must be available to use fused_adam.")

    optimizer = AVAILABLE_OPTIMIZERS[name]
    optimizer = partial(optimizer, **kwargs)
    return optimizer



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/unetr.py

from typing import Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch import nn

from atommic.collections.segmentation.nn.vit_base.utils import get_conv_layer
from atommic.collections.segmentation.nn.vit_base.vit_block import ViT


class UnetOutBlock(nn.Module):
    """Implementation of the output block of UNETR, as presented in [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).

    .. note::
        This is a wrapper for monai implementation of UNETR.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/unetr.py
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        dropout: Optional[Union[Tuple, str, float]] = None,
    ):
        """Inits :class:`UnetOutBlock`.

        Parameters
        ----------
        spatial_dims : int
            Number of spatial dimensions of the input image.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        dropout : Optional[Union[Tuple, str, float]]
            Dropout rate.
        """
        super().__init__()
        self.conv = get_conv_layer(
            spatial_dims,
            in_channels,
            out_channels,
            kernel_size=1,
            stride=1,
            dropout=dropout,
            bias=True,
            act=None,
            norm=None,
            conv_only=False,
        )

    def forward(self, inp):
        """Forward pass of :class:`UnetOutBlock`."""
        return self.conv(inp)


class UnetrBasicBlock(nn.Module):
    """A CNN module that can be used for UNETR, as presented in [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).

    .. note::
        This is a wrapper for monai implementation of UNETR.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/unetr.py
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[Sequence[int], int],
        stride: Union[Sequence[int], int],
        norm_name: Union[Tuple, str],
        res_block: bool = False,
    ):
        """Inits :class:`UnetrBasicBlock`.

        Parameters
        ----------
        spatial_dims : int
            Number of spatial dimensions of the input image.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        kernel_size : Union[Sequence[int], int]
            Convolution kernel size.
        stride : Union[Sequence[int], int]
            Convolution stride.
        norm_name : Union[Tuple, str]
            Feature normalization type and arguments.
        res_block : bool
            If True, use a residual block.
        """
        super().__init__()

        if res_block:
            self.layer = UnetResBlock(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                norm_name=norm_name,
            )
        else:
            self.layer = UnetBasicBlock(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                norm_name=norm_name,
            )

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`UnetrBasicBlock`."""
        return self.layer(inp)


class UnetrPrUpBlock(nn.Module):
    """A projection upsampling module that can be used for UNETR, as presented in [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).

    .. note::
        This is a wrapper for monai implementation of UNETR.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/unetr.py
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        num_layer: int,
        kernel_size: Union[Sequence[int], int],
        stride: Union[Sequence[int], int],
        upsample_kernel_size: Union[Sequence[int], int],
        norm_name: Union[Tuple, str],
        conv_block: bool = False,
        res_block: bool = False,
    ):
        """Inits :class:`UnetrPrUpBlock`.

        Parameters
        ----------
        spatial_dims : int
            Number of spatial dimensions of the input image.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        num_layer : int
            Number of layers.
        kernel_size : Union[Sequence[int], int]
            Convolution kernel size.
        stride : Union[Sequence[int], int]
            Convolution stride.
        upsample_kernel_size : Union[Sequence[int], int]
            Upsampling kernel size.
        norm_name : Union[Tuple, str]
            Feature normalization type and arguments.
        conv_block : bool
            If True, use a convolution block.
        res_block : bool
            If True, use a residual block.
        """
        super().__init__()

        upsample_stride = upsample_kernel_size
        self.transp_conv_init = get_conv_layer(
            spatial_dims,
            in_channels,
            out_channels,
            kernel_size=upsample_kernel_size,
            stride=upsample_stride,
            conv_only=True,
            is_transposed=True,
        )
        if conv_block:
            if res_block:
                self.blocks = nn.ModuleList(
                    [
                        nn.Sequential(
                            get_conv_layer(
                                spatial_dims,
                                out_channels,
                                out_channels,
                                kernel_size=upsample_kernel_size,
                                stride=upsample_stride,
                                conv_only=True,
                                is_transposed=True,
                            ),
                            UnetResBlock(
                                spatial_dims=spatial_dims,
                                in_channels=out_channels,
                                out_channels=out_channels,
                                kernel_size=kernel_size,
                                stride=stride,
                                norm_name=norm_name,
                            ),
                        )
                        for i in range(num_layer)
                    ]
                )
            else:
                self.blocks = nn.ModuleList(
                    [
                        nn.Sequential(
                            get_conv_layer(
                                spatial_dims,
                                out_channels,
                                out_channels,
                                kernel_size=upsample_kernel_size,
                                stride=upsample_stride,
                                conv_only=True,
                                is_transposed=True,
                            ),
                            UnetBasicBlock(
                                spatial_dims=spatial_dims,
                                in_channels=out_channels,
                                out_channels=out_channels,
                                kernel_size=kernel_size,
                                stride=stride,
                                norm_name=norm_name,
                            ),
                        )
                        for i in range(num_layer)
                    ]
                )
        else:
            self.blocks = nn.ModuleList(
                [
                    get_conv_layer(
                        spatial_dims,
                        out_channels,
                        out_channels,
                        kernel_size=upsample_kernel_size,
                        stride=upsample_stride,
                        conv_only=True,
                        is_transposed=True,
                    )
                    for i in range(num_layer)
                ]
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`UnetrPrUpBlock`."""
        x = self.transp_conv_init(x)
        for blk in self.blocks:
            x = blk(x)
        return x


class UnetrUpBlock(nn.Module):
    """An upsampling module that can be used for UNETR, as presented in [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).

    .. note::
        This is a wrapper for monai implementation of UNETR.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/unetr.py
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[Sequence[int], int],
        upsample_kernel_size: Union[Sequence[int], int],
        norm_name: Union[Tuple, str],
        res_block: bool = False,
    ):
        """Inits :class:`UnetrUpBlock`.

        Parameters
        ----------
        spatial_dims : int
            Number of spatial dimensions of the input image.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        kernel_size : Union[Sequence[int], int]
            Convolution kernel size.
        upsample_kernel_size : Union[Sequence[int], int]
            Upsampling kernel size.
        norm_name : Union[Tuple, str]
            Feature normalization type and arguments.
        res_block : bool
            If True, use a residual block.
        """
        super().__init__()
        upsample_stride = upsample_kernel_size
        self.transp_conv = get_conv_layer(
            spatial_dims,
            in_channels,
            out_channels,
            kernel_size=upsample_kernel_size,
            stride=upsample_stride,
            conv_only=True,
            is_transposed=True,
        )

        if res_block:
            self.conv_block = UnetResBlock(
                spatial_dims,
                out_channels + out_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=1,
                norm_name=norm_name,
            )
        else:
            self.conv_block = UnetBasicBlock(
                spatial_dims,
                out_channels + out_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=1,
                norm_name=norm_name,
            )

    def forward(self, inp: torch.Tensor, skip: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`UnetrUpBlock`."""
        # number of channels for skip should equals to out_channels
        return self.conv_block(torch.cat((self.transp_conv(inp), skip), dim=1))


class UnetResBlock(nn.Module):
    """A skip-connection based module for UNETR, as presented in [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[Sequence[int], int],
        stride: Union[Sequence[int], int],
        norm_name: Union[Tuple, str],  # pylint: disable=unused-argument
        act_name: Union[Tuple, str] = (  # pylint: disable=unused-argument
            "leakyrelu",
            {"inplace": True, "negative_slope": 0.01},
        ),
        dropout: Optional[Union[Tuple, str, float]] = None,
    ):
        """Inits :class:`UnetResBlock`.

        Parameters
        ----------
        spatial_dims : int
            Number of spatial dimensions of the input image.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        kernel_size : Union[Sequence[int], int]
            Convolution kernel size.
        stride : Union[Sequence[int], int]
            Convolution stride.
        norm_name : Union[Tuple, str]
            Feature normalization type and arguments.
        act_name : Union[Tuple, str]
            Activation function type and arguments.
        dropout : Optional[Union[Tuple, str, float]]
            Dropout rate.
        """
        super().__init__()
        self.conv1 = get_conv_layer(
            spatial_dims,
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            dropout=dropout,
            act=None,
            norm=None,
            conv_only=False,
        )
        self.conv2 = get_conv_layer(
            spatial_dims,
            out_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=1,
            dropout=dropout,
            act=None,
            norm=None,
            conv_only=False,
        )
        self.lrelu = nn.LeakyReLU(negative_slope=0.01, inplace=True)
        if spatial_dims == 2:
            self.norm1 = nn.InstanceNorm2d(out_channels)
            self.norm2 = nn.InstanceNorm2d(out_channels)
            self.norm3 = nn.InstanceNorm2d(out_channels)
        elif spatial_dims == 3:
            self.norm1 = nn.InstanceNorm3d(out_channels)
            self.norm2 = nn.InstanceNorm3d(out_channels)
            self.norm3 = nn.InstanceNorm3d(out_channels)
        self.downsample = in_channels != out_channels
        stride_np = np.atleast_1d(stride)
        if not np.all(stride_np == 1):
            self.downsample = True
        if self.downsample:
            self.conv3 = get_conv_layer(
                spatial_dims,
                in_channels,
                out_channels,
                kernel_size=1,
                stride=stride,
                dropout=dropout,
                act=None,
                norm=None,
                conv_only=False,
            )

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`UnetResBlock`."""
        residual = inp
        out = self.conv1(inp)
        out = self.norm1(out)
        out = self.lrelu(out)
        out = self.conv2(out)
        out = self.norm2(out)
        if hasattr(self, "conv3"):
            residual = self.conv3(residual)
        if hasattr(self, "norm3"):
            residual = self.norm3(residual)
        out = out + residual
        out = self.lrelu(out)
        return out


class UnetUpBlock(nn.Module):
    """An upsampling module that can be used for UNETR, as presented in [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).

    .. note::
        This is a wrapper for monai implementation of UNETR.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/unetr.py
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[Sequence[int], int],
        stride: Union[Sequence[int], int],  # pylint: disable=unused-argument
        upsample_kernel_size: Union[Sequence[int], int],
        norm_name: Union[Tuple, str],
        act_name: Union[Tuple, str] = ("leakyrelu", {"inplace": True, "negative_slope": 0.01}),
        dropout: Optional[Union[Tuple, str, float]] = None,
        trans_bias: bool = False,
    ):
        """Inits :class:`UnetUpBlock`.

        Parameters
        ----------
        spatial_dims : int
            Number of spatial dimensions of the input image.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        kernel_size : Union[Sequence[int], int]
            Convolution kernel size.
        stride : Union[Sequence[int], int]
            Convolution stride.
        upsample_kernel_size : Union[Sequence[int], int]
            Upsampling kernel size.
        norm_name : Union[Tuple, str]
            Feature normalization type and arguments.
        act_name : Union[Tuple, str]
            Activation function type and arguments.
        dropout : Optional[Union[Tuple, str, float]]
            Dropout rate.
        trans_bias : bool
            Whether to use bias in the transposed convolution layer. Default is ``False``.
        """
        super().__init__()
        upsample_stride = upsample_kernel_size
        self.transp_conv = get_conv_layer(
            spatial_dims,
            in_channels,
            out_channels,
            kernel_size=upsample_kernel_size,
            stride=upsample_stride,
            dropout=dropout,
            bias=trans_bias,
            act=None,
            norm=None,
            conv_only=False,
            is_transposed=True,
        )
        self.conv_block = UnetBasicBlock(
            spatial_dims,
            out_channels + out_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=1,
            dropout=dropout,
            norm_name=norm_name,
            act_name=act_name,
        )

    def forward(self, inp: torch.Tensor, skip: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`UnetUpBlock`."""
        # number of channels for skip should equal to out_channels
        return self.conv_block(torch.cat((self.transp_conv(inp), skip), dim=1))


class UnetBasicBlock(nn.Module):
    """A CNN module that can be used for UNETR, as presented in  [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).

    .. note::
        This is a wrapper for monai implementation of UNETR.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/unetr.py
    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[Sequence[int], int],
        stride: Union[Sequence[int], int],
        norm_name: Union[Tuple, str],  # pylint: disable=unused-argument
        act_name: Union[Tuple, str] = (  # pylint: disable=unused-argument
            "leakyrelu",
            {"inplace": True, "negative_slope": 0.01},
        ),
        dropout: Optional[Union[Tuple, str, float]] = None,
    ):
        """Inits :class:`UnetBasicBlock`.

        Parameters
        ----------
        spatial_dims : int
            Number of spatial dimensions of the input image.
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        kernel_size : Union[Sequence[int], int]
            Convolution kernel size.
        stride : Union[Sequence[int], int]
            Convolution stride.
        norm_name : Union[Tuple, str]
            Feature normalization type and arguments.
        act_name : Union[Tuple, str]
            Activation function type and arguments.
        dropout : Optional[Union[Tuple, str, float]]
            Dropout rate.
        """
        super().__init__()
        self.conv1 = get_conv_layer(
            spatial_dims,
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            dropout=dropout,
            act=None,
            norm=None,
            conv_only=False,
        )
        self.conv2 = get_conv_layer(
            spatial_dims,
            out_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=1,
            dropout=dropout,
            act=None,
            norm=None,
            conv_only=False,
        )
        self.lrelu = nn.LeakyReLU(negative_slope=0.01, inplace=True)
        if spatial_dims == 2:
            self.norm1 = nn.InstanceNorm2d(out_channels)
            self.norm2 = nn.InstanceNorm2d(out_channels)
        elif spatial_dims == 3:
            self.norm1 = nn.InstanceNorm3d(out_channels)
            self.norm2 = nn.InstanceNorm3d(out_channels)

    def forward(self, inp: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`UnetBasicBlock`."""
        return self.lrelu(self.norm2(self.conv2(self.lrelu(self.norm1(self.conv1(inp))))))


class UNETR(nn.Module):
    """UNETR as presented in [Hatamizadeh2022]_.

    References
    ----------
    .. [Hatamizadeh2022] Hatamizadeh A, Tang Y, Nath V, Yang D, Myronenko A, Landman B, Roth HR, Xu D. Unetr:
        Transformers for 3d medical image segmentation. InProceedings of the IEEE/CVF Winter Conference on
        Applications of Computer Vision 2022 (pp. 574-584).

    .. note::
        This is a wrapper for monai implementation of UNETR.
        See: https://github.com/Project-MONAI/MONAI/blob/dev/monai/networks/nets/unetr.py
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        img_size: Union[Sequence[int], int],
        feature_size: int = 16,
        hidden_size: int = 768,
        mlp_dim: int = 3072,
        num_heads: int = 12,
        pos_embed: str = "conv",
        norm_name: Union[Tuple, str] = "instance",
        conv_block: bool = True,
        res_block: bool = True,
        dropout_rate: float = 0.0,
        spatial_dims: int = 3,
        qkv_bias: bool = False,
    ):
        """Inits :class:`UNETR`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        img_size : Union[Sequence[int], int]
            Dimension of input image.
        feature_size : int
            Dimension of network feature size. Default is ``16``.
        hidden_size : int
            Dimension of network hidden size. Default is ``768``.
        mlp_dim : int
            Dimension of network mlp size. Default is ``3072``.
        num_heads : int
            Number of attention heads. Default is ``12``.
        pos_embed : str
            Positional embedding type. Default is ``"conv"``.
        norm_name : Union[Tuple, str]
            Feature normalization type and arguments. Default is ``"instance"``.
        conv_block : bool
            Whether to use convolutional block. Default is ``True``.
        res_block : bool
            Whether to use residual block. Default is ``True``.
        dropout_rate : float
            Dropout rate. Default is ``0.0``.
        spatial_dims : int
            Number of spatial dimensions of the input image. Default is ``3``.
        qkv_bias : bool
            Whether to use bias for qkv. Default is ``False``.
        """
        super().__init__()
        if not 0 <= dropout_rate <= 1:
            raise ValueError("dropout_rate should be between 0 and 1.")
        if hidden_size % num_heads != 0:
            raise ValueError("hidden_size should be divisible by num_heads.")

        self.num_layers = 12
        self.patch_size = (16,) * spatial_dims
        self.feat_size = tuple(img_d // p_d for img_d, p_d in zip(img_size, self.patch_size))  # type: ignore
        self.hidden_size = hidden_size
        self.classification = False
        self.vit = ViT(
            in_channels=in_channels,
            img_size=img_size,
            patch_size=self.patch_size,
            hidden_size=hidden_size,
            mlp_dim=mlp_dim,
            num_layers=self.num_layers,
            num_heads=num_heads,
            pos_embed=pos_embed,
            classification=self.classification,
            dropout_rate=dropout_rate,
            spatial_dims=spatial_dims,
            qkv_bias=qkv_bias,
        )
        self.encoder1 = UnetrBasicBlock(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            out_channels=feature_size,
            kernel_size=3,
            stride=1,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.encoder2 = UnetrPrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=hidden_size,
            out_channels=feature_size * 2,
            num_layer=2,
            kernel_size=3,
            stride=1,
            upsample_kernel_size=2,
            norm_name=norm_name,
            conv_block=conv_block,
            res_block=res_block,
        )
        self.encoder3 = UnetrPrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=hidden_size,
            out_channels=feature_size * 4,
            num_layer=1,
            kernel_size=3,
            stride=1,
            upsample_kernel_size=2,
            norm_name=norm_name,
            conv_block=conv_block,
            res_block=res_block,
        )
        self.encoder4 = UnetrPrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=hidden_size,
            out_channels=feature_size * 8,
            num_layer=0,
            kernel_size=3,
            stride=1,
            upsample_kernel_size=2,
            norm_name=norm_name,
            conv_block=conv_block,
            res_block=res_block,
        )
        self.decoder5 = UnetrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=hidden_size,
            out_channels=feature_size * 8,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.decoder4 = UnetrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=feature_size * 8,
            out_channels=feature_size * 4,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.decoder3 = UnetrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=feature_size * 4,
            out_channels=feature_size * 2,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.decoder2 = UnetrUpBlock(
            spatial_dims=spatial_dims,
            in_channels=feature_size * 2,
            out_channels=feature_size,
            kernel_size=3,
            upsample_kernel_size=2,
            norm_name=norm_name,
            res_block=res_block,
        )
        self.out = UnetOutBlock(spatial_dims=spatial_dims, in_channels=feature_size, out_channels=out_channels)
        self.proj_axes = (0, spatial_dims + 1) + tuple(d + 1 for d in range(spatial_dims))
        self.proj_view_shape = list(self.feat_size) + [self.hidden_size]

    def proj_feat(self, x: torch.Tensor) -> torch.Tensor:
        """Project the feature map to the hidden size."""
        new_view = [x.size(0)] + self.proj_view_shape
        x = x.view(new_view)
        x = x.permute(self.proj_axes).contiguous()
        return x

    def forward(self, x_in: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`UNETR`."""
        x, hidden_states_out = self.vit(x_in)
        enc1 = self.encoder1(x_in)
        x2 = hidden_states_out[3]
        enc2 = self.encoder2(self.proj_feat(x2))
        x3 = hidden_states_out[6]
        enc3 = self.encoder3(self.proj_feat(x3))
        x4 = hidden_states_out[9]
        enc4 = self.encoder4(self.proj_feat(x4))
        dec4 = self.proj_feat(x)
        dec3 = self.decoder5(dec4, enc4)
        dec2 = self.decoder4(dec3, enc3)
        dec1 = self.decoder3(dec2, enc2)
        out = self.decoder2(dec1, enc1)
        return self.out(out)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.reconstruction.losses.na import NoiseAwareLoss  # noqa: F401
from atommic.collections.reconstruction.losses.ssim import SSIMLoss  # noqa: F401



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from:
# https://github.com/NVIDIA/NeMo/blob/main/nemo/collections/common/metrics/global_average_loss_metric.py

import torch
from torchmetrics import Metric

__all__ = ["GlobalAverageLossMetric"]


class GlobalAverageLossMetric(Metric):
    """This class is for averaging loss across multiple processes if a distributed backend is used. True average is
    computed not running average. It does not accumulate gradients so the averaged loss cannot be used for
    optimization.

    .. note::
        If ``take_avg_loss`` is ``True``, the :meth:`update` method ``loss`` argument has to be a mean loss. If
        ``take_avg_loss`` is ``False`` then the :meth:`update` method ``loss`` argument has to be a sum of losses. See
        PyTorch Lightning Metrics for the metric usage instruction.

    Examples
    --------
    >>> from atommic.collections.common.metrics.global_average_loss_metric import GlobalAverageLossMetric
    >>> metric = GlobalAverageLossMetric()
    >>> metric.update(torch.tensor(1.0), torch.tensor(1))
    >>> metric.update(torch.tensor(2.0), torch.tensor(1))
    >>> metric.compute()
    tensor(1.5000)
    >>> metric.update(torch.tensor(3.0), torch.tensor(1))
    >>> metric.compute()
    tensor(2.0000)
    """

    full_state_update: bool = True

    def __init__(self, dist_sync_on_step=False, process_group=None, take_avg_loss=True):
        """Inits :class:`GlobalAverageLossMetric`.

        Parameters
        ----------
        dist_sync_on_step : bool
            Synchronize metric state across processes at each method :meth:`forward` call before returning the value at
             the step. Default is ``False``.
        process_group : Any, optional
            Specify the process group on which synchronization is called. default: ``None`` (which selects the entire
            world). Default is ``None``.
        take_avg_loss : bool
            If ``True`` values of :meth:`update` method ``loss`` argument has to be a mean loss. If ``False`` values of
            :meth:`update` method ``loss`` argument has to be a sum of losses. Default is ``True``.
        """
        super().__init__(dist_sync_on_step=dist_sync_on_step, process_group=process_group)
        self.add_state("loss_sum", torch.tensor(0.0, dtype=torch.float64), dist_reduce_fx='sum')
        self.add_state("num_measurements", torch.tensor(0, dtype=torch.int64), dist_reduce_fx='sum')
        self.take_avg_loss = take_avg_loss

    def update(self, loss, num_measurements):  # pylint: disable=arguments-differ
        """Updates :attr:`loss_sum` and :attr:`num_measurements`.

        Parameters
        ----------
        loss : torch.Tensor
            A float zero dimensional ``torch.Tensor`` which is either sum or average of losses for processed examples.
            See ``take_avg_loss`` parameter of :meth:`__init__`.
        num_measurements : torch.Tensor
            An integer zero dimensional ``torch.Tensor`` which contains a number of loss measurements. The sum or mean
            of the results of these measurements are in the ``loss`` parameter.
        """
        if self.take_avg_loss:
            self.loss_sum = self.loss_sum + loss.detach() * num_measurements
        else:
            self.loss_sum = self.loss_sum + loss.detach()
        self.num_measurements = self.num_measurements + num_measurements

    def compute(self):
        """Returns mean loss."""
        if self.num_measurements.eq(0):
            return torch.tensor(float("nan"))
        return self.loss_sum / self.num_measurements
